/*
** Module   : FrameImpl.java
** Abstract : Frame implementation for CHUI
**
** Copyright (c) 2011-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- --------------------------Description-----------------------------
** 001 SIY 20111121          Extracted CHUI-specific part from Frame.
** 002 EVL 20140514          Fix for widget underlining for still non-visible frames. This case
**                           requires one time per frame shift for down body content one line up.
*/
 
package com.goldencode.p2j.ui.chui;

import java.util.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;


/**
 * A <code>Frame</code> is a top-level window with a title and a border. 
 */
@SuppressWarnings("unchecked")
public class FrameImpl
extends Frame
{
   /**
    * Special constructor used to restore component using given configuration.
    * 
    * @param   cfg
    *          Configuration data.
    */
   public FrameImpl(ComponentConfig cfg)
   {
      super(cfg);
   }
   
   /**
    * Draw this container.
    */
   @Override
   public void draw()
   {
      // skip drawing if current system mode does not match frame mode 
      if (redirected != ThinClient.getInstance().isRedirected())
         return;
      
      if (isRedirected() != isPlacedForRedirected())
      {
         // frame placement does not match current frame state (redirected or
         // interactive) so we can't draw frame correctly and ought to ignore
         // the request
         return;
      }
      
      // draw the enclosing frame (but only if the insets are nonzero)
      Point origReal = new Point(location());
      Dimension sizeReal = size().copy();
      
      sizeReal.width -= calculateWidthDelta();

      if (sizeReal.height <= 0)
         return;

      // In interactive mode drawing of empty frames is suppressed.
      // Also, since such a frame is not displayed, it should not affect
      // pausing.
      if (!isRedirected() &&  sizeReal.width <= 0)
      {
         needPause = false;
         return;
      }
      
      draw(origReal, sizeReal);
      
      chooseHandler.drawIfActive();
   }

   /**
    * Draw the peer container. The methods calls the draw() method for
    * ScrollPane instance for the peer container. In addition the required
    * clipping rectangle is calculated for ChildFrame containers because they
    * must be rendered inside the parent container.
    * <p>
    * Drawing for the redirected terminal differs. Labels/headers are drawn
    * only once per page, instead of the usual once per frame.
    * 
    * @param   origin
    *          The coordinate of the upper-left corner for peer container in
    *          character units.
    * @param   size
    *          The size of the peer container in character units.
    */
   public void draw(Point origin, Dimension size)
   {
      ThinClient tc = ThinClient.getInstance();
      
      boolean drawTitle = false;
      boolean needSync = false;
      boolean box = config().isBox();
      boolean redir = tc.isRedirected();
      Point z = contentPane.screenLocation();
      z.y += headersHeight;
      
      Color tcolor = config().getTitleDcolor();
      
      // Getting the current rectangle
      Rectangle thisRect = new Rectangle(origin,
                                         new Dimension(size.width + 1,
                                                       size.height + 1));
      origin.x += config().getBorderLeftChars(); 
      
      if (tc.viewRefresh && !redir)
      {
         screen().clear(origin, size, UiUtils.getDcolor(this));
         needSync = true;
      }

      if (frameScroll != null)
      {
         if (!down)
         {
            frameScroll.draw();
            
            if (redir)
            {
               if (!box)
               {
                  // force leading skips
                  forceSkips(thisRect.top()); 
               }
               
               if (trailingSkips)
                  forceSkips(thisRect.bottom() - 1);

               Stream stream = tc.getCurrentRedirection();
               RedirectedTerminal rt = 
                  ThinClient.redirectedTerminalForStream(stream);
               rt.markHeader(headersHeight);
               drawTitle = true;
            }
            
            needSync = true;
         }
         else
         {
            // control frame decorations/headers drawing for interactive and
            // redirected cases
            if (redir)
            {
               Stream stream = tc.getCurrentRedirection();
               
               if (headersOutput == null)
               {
                  headersOutput = new HashSet<Stream>();
               }
               
               // test to see if we have output yet on this page
               if (stream != null && !headersOutput.contains(stream))
               {
                  RedirectedTerminal rt = 
                              ThinClient.redirectedTerminalForStream(stream);
                  
                  // remember that we have output already, this will be
                  // reset based on notifications
                  headersOutput.add(stream);
                  if (stream.rawPageSize() > 0)
                  {
                     stream.registerPageBreakListener(this);
                  }
                  else
                  {
                     stream.registerCloseListener(this);
                  }
                  
                  if (box)
                  {
                     // all frames generate a line at the top of a new page
                     // preserve the syncable status 
                     forceNewLine();
                     rt.flush();
                     
                     // if we have headers, mark those lines as boxed
                     if (headersHeight > 0)
                        forceBox(z.y - headersHeight, headersHeight);
                  }
                  
                  // draw the headers
                  frameScroll.draw(nonbody);
                  
                  // the following sync call is required so that the next
                  // iteration does not cause adding blank lines between the
                  // header labels and that iteration!
                  screen().sync();
                  needSync = false;

                  rt.markHeader(headersHeight);
                  
                  drawTitle = true;
               }
            }
            else if (tc.viewRefresh)
            {   
               frameScroll.draw(nonbody);
               needSync = true;
            }
            
            int from = 0;
            int to = 0;
            
            if (tc.viewDisplay || redir)
            {
               from = cursor >= 0 ? cursor : 0;
               to = from + 1;
            }
            else
            {
               from = 0;
               to = downBody.length;
            }
            
            Dimension dim = frameScroll.minimumSize();
            dim.height -= config().getBorderTopChars() +
                          config().getBorderBottomChars();
            int row = getFixedHeight();
            
            Rectangle boundsRect = frameScroll.bounds();
            
            // in case we underline when there is no frame on the screen we need to move
            // all widgets one line up to compensate down shift on the server side
            if (isPendingUnderline())
            {
               setPendingUnderline(false);
               for (int i = from; i < to; i++)
               {
                  for (int j = 0; j < downBody[i].length; j++)
                  {
                     relocateComponent(downBody[i][j], -1, downBody[i][j]);
                  }
               }
            }
            
            for (int i = from; i < to; i++, row += getIterationHeight())
            {
               if (!touched[i])
               {
                  continue;
               }
               
               if (!redir && row >= dim.height)
                  break;
               
               if (redir && box)
               {
                  int k =  underlinedRow[i] ? 1 : verticalStepBoxed;
                  forceBox(z.y + startRow[0], k);
                  needSync = true;
               }
               
               for (int j = 0; j < downBody[i].length; j++)
               {
                  if (underlinedRow[i] || i > lastRow)
                  {
                     if (underline[i][j])
                     {
                        drawUnderline(downBody[i][j]);
                        if (redir && i==0 && downBody[i][j] instanceof FillIn &&
                            ((FillIn)downBody[i][j]).getValue() instanceof decimal &&
                            ((decimal)((FillIn)downBody[i][j]).getValue()).equals(new decimal("0")) )
                        {
                           relocateComponent(downBody[i][j], 1, downBody[i][j]);
                           downBody[i][j].draw();
                        }
                           
                        needSync = true;
                     }
                     
                     continue;
                  }
                  
                  // clip components which are outside bounds
                  if (!redir &&
                      !boundsRect.contains(downBody[i][j].bounds()))
                  {
                     break;
                  }
                  
                  downBody[i][j].draw();
                  needSync = true;
                  
                  if (downBody[i][j] instanceof LabeledWidget && 
                      isSideLabels())
                  {
                     LabeledWidget lw = (LabeledWidget) downBody[i][j];
                     ComponentConfig cfg = downBody[i][j].config();
                     
                     if (!lw.isNoLabels() && 
                         lw.getLabelInstance() != null &&
                         cfg.getState() != ScreenBuffer.UNINITIALIZED)
                     {
                        lw.getLabelInstance().draw();
                        needSync = true;
                     }
                  }
               }
            }
         }
      }

      if (((tc.viewRefresh && !redir) || drawTitle) && hasTitle())
      {
         // Draw the title into the enclosing frame.
         drawTitle(origin, size.width, redir, tcolor);
      }
      
      if (needSync)
         screen().sync();
   }

   /**
    * Draw underline for specified widget.
    * 
    * @param   comp
    *          The base component to draw underline for.
    */
   private void drawUnderline(Widget comp)
   {
      Point p = comp.screenLocation();
      
      screen().at(p);
  
      Color color = UiUtils.getDcolor(this).withoutReverse();

      // underline does not fit into current rectangle
      ScreenBitmap bitmap = screen().getAndResetBitmap();
      
      screen().hLine(comp.width(), color).setBitmap(bitmap);
   }
   
   /**
    * Draw frame title.
    * 
    * @param   origin
    *          Frame origin. 
    * @param   width
    *          Frame width.
    * @param   redir
    *          Redirected mode flag.
    * @param   tcolor
    *          Title color.
    */
   private void drawTitle(Point origin, int width, boolean redir, Color tcolor)
   {
      Point p = new Point(origin);

      if (redir)
         p.x += (width - config().getTitle().length() + 1)/2 - 1;
      else
         p.x += (width - config().getTitle().length())/2;

      screen().at(p).append(config().getTitle(), tcolor);
   }
   
   /**
    * Draw all uninitialized widgets of the current iteration.
    */
   @Override
   public void drawUninitialized()
   {
      int i = cursor;
      
      if (i < 0 || downBody == null || downBody.length == 0)
      {
         i = 0;
      }
      
      if (i > lastRow || underlinedRow[i] || !touched[i])
         return;
      
      for (int j = 0; j < downBody[i].length; j++)
      {
         if (downBody[i][j].config() == null)
            continue;

         ComponentConfig cfg = downBody[i][j].config();
         if (cfg.getState() != ScreenBuffer.UNINITIALIZED)
            continue;
         
         downBody[i][j].draw();
      }
   }
}
